@@ -12,9 +12,9 @@ type KuryrBootstrapResult struct {
 	PodSecurityGroups        []string
 	ExternalNetwork          string
 	ClusterID                string
+	OctaviaProvider          string
 	OctaviaMultipleListeners bool
 	OpenStackCloud           clientconfig.Cloud
-	OctaviaVersion           string
 	WebhookCA                string
 	WebhookCAKey             string
 	WebhookCert              string
