@@ -15,7 +15,6 @@ import (
 	"sigs.k8s.io/controller-runtime/pkg/client"
 	"strings"
 
-	"github.com/Masterminds/semver"
 	"github.com/pkg/errors"
 	"gopkg.in/yaml.v2"
 
@@ -26,6 +25,7 @@ import (
 	"github.com/gophercloud/gophercloud/openstack/loadbalancer/v2/loadbalancers"
 	"github.com/gophercloud/gophercloud/openstack/loadbalancer/v2/monitors"
 	"github.com/gophercloud/gophercloud/openstack/loadbalancer/v2/pools"
+	"github.com/gophercloud/gophercloud/openstack/loadbalancer/v2/providers"
 	"github.com/gophercloud/gophercloud/openstack/networking/v2/extensions/attributestags"
 	"github.com/gophercloud/gophercloud/openstack/networking/v2/extensions/layer3/routers"
 	"github.com/gophercloud/gophercloud/openstack/networking/v2/extensions/security/groups"
@@ -38,7 +38,6 @@ import (
 	"github.com/openshift/cluster-network-operator/pkg/names"
 	"github.com/openshift/cluster-network-operator/pkg/platform/openstack/util/cert"
 
-	apierrors "k8s.io/apimachinery/pkg/api/errors"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
 
 	confv1 "github.com/openshift/api/config/v1"
@@ -57,20 +56,20 @@ const (
 	InfrastructureCRDName                  = "cluster"
 	MinOctaviaVersionWithMultipleListeners = "v2.11"
 	MinOctaviaVersionWithHTTPSMonitors     = "v2.10"
+	MinOctaviaVersionWithProviders         = "v2.6"
 	MinOctaviaVersionWithTagSupport        = "v2.5"
 	MinOctaviaVersionWithTimeouts          = "v2.1"
 	KuryrNamespace                         = "openshift-kuryr"
 	KuryrConfigMapName                     = "kuryr-config"
-	DNSNamespace                           = "openshift-dns"
-	DNSServiceName                         = "dns-default"
-	etcdClientPort                         = 2379
-	etcdServerToServerPort                 = 2380
-	dnsPort                                = 53
-	apiPort                                = 6443
-	routerMetricsPort                      = 1936
-	kubeCMMetricsPort                      = 10257
-	kubeSchedulerMetricsPort               = 10259
-	kubeletMetricsPort                     = 10250
+	// NOTE(ltomasbo): Only OVN octavia driver supported on kuryr
+	OVNProvider              = "ovn"
+	etcdPort                 = 2379
+	dnsPort                  = 53
+	apiPort                  = 6443
+	routerMetricsPort        = 1936
+	kubeCMMetricsPort        = 10257
+	kubeSchedulerMetricsPort = 10259
+	kubeletMetricsPort       = 10250
 )
 
 func GetClusterID(kubeClient client.Client) (string, error) {
@@ -447,8 +446,8 @@ func ensureOpenStackSgRule(client *gophercloud.ServiceClient, sgId, remotePrefix
 		Direction:      rules.DirIngress,
 		RemoteIPPrefix: remotePrefix,
 	}
-	// Let's just assume that we're getting passed 0 when we aren't supposed to set those
-	if portMin > 0 && portMax > 0 {
+	// Let's just assume that we're getting passed -1 when we aren't supposed to set those
+	if portMin >= 0 && portMax >= 0 {
 		opts.PortRangeMin = portMin
 		opts.PortRangeMax = portMax
 		opts.Protocol = protocol
@@ -464,29 +463,6 @@ func ensureOpenStackSgRule(client *gophercloud.ServiceClient, sgId, remotePrefix
 	return nil
 }
 
-// Returns list of OpenStack ingress security group rules on SGs tagged with tag.
-func listOpenStackSgRules(client *gophercloud.ServiceClient, tag string) ([]rules.SecGroupRule, error) {
-	page, err := groups.List(client, groups.ListOpts{Tags: tag}).AllPages()
-	if err != nil {
-		return nil, errors.Wrap(err, "failed to get security group list")
-	}
-	groupsList, err := groups.ExtractGroups(page)
-	if err != nil {
-		return nil, errors.Wrap(err, "failed to extract security group list")
-	}
-
-	var rulesList []rules.SecGroupRule
-	for _, group := range groupsList {
-		for _, rule := range group.Rules {
-			if rule.Direction == string(rules.DirIngress) {
-				rulesList = append(rulesList, rule)
-			}
-		}
-	}
-
-	return rulesList, nil
-}
-
 // Delete a LoadBalancer and all associated resources
 func deleteOpenStackLb(client *gophercloud.ServiceClient, lbId string) error {
 	log.Printf("Deleting Openstack LoadBalancer: %s", lbId)
@@ -516,7 +492,8 @@ func deleteOpenStackLb(client *gophercloud.ServiceClient, lbId string) error {
 }
 
 // Waits up to 5 minutes for OpenStack LoadBalancer to move into ACTIVE
-// provisioning_status. Fails if time runs out.
+// provisioning_status. Fails if time runs out, or LoadBalancer goes in ERROR
+// state.
 func waitForOpenStackLb(client *gophercloud.ServiceClient, lbId string) error {
 	err := gophercloud.WaitFor(300, func() (bool, error) {
 		lb, err := loadbalancers.Get(client, lbId).Extract()
@@ -524,6 +501,10 @@ func waitForOpenStackLb(client *gophercloud.ServiceClient, lbId string) error {
 			return false, err
 		}
 
+		if lb.ProvisioningStatus == "ERROR" {
+			return true, errors.Errorf("LoadBalancer gone in error state")
+		}
+
 		return lb.ProvisioningStatus == "ACTIVE", nil
 	})
 
@@ -621,7 +602,7 @@ func ensureOpenStackLb(client *gophercloud.ServiceClient, name, vipAddress, vipS
 	}
 	err = waitForOpenStackLb(client, lb.ID)
 	if err != nil {
-		return "", errors.Errorf("Timed out waiting for the LB %s to become ready", lb.ID)
+		return "", errors.Wrapf(err, "Error waiting for LB %s", lb.ID)
 	}
 
 	return lb.ID, nil
@@ -661,7 +642,7 @@ func ensureOpenStackLbPool(client *gophercloud.ServiceClient, name, lbId string)
 
 		err = waitForOpenStackLb(client, lbId)
 		if err != nil {
-			return "", errors.Errorf("Timed out waiting for the LB %s to become ready", lbId)
+			return "", errors.Wrapf(err, "Error waiting for LB %s", lbId)
 		}
 
 		return poolsObj.ID, nil
@@ -754,7 +735,7 @@ func ensureOpenStackLbPoolMember(client *gophercloud.ServiceClient, name, lbId,
 
 		err = waitForOpenStackLb(client, lbId)
 		if err != nil {
-			return "", errors.Errorf("Timed out waiting for the LB %s to become ready", lbId)
+			return "", errors.Wrapf(err, "Error waiting for LB %s", lbId)
 		}
 
 		return poolsObj.ID, nil
@@ -811,7 +792,7 @@ func ensureOpenStackLbListener(client *gophercloud.ServiceClient, name, lbId, po
 
 		err = waitForOpenStackLb(client, lbId)
 		if err != nil {
-			return "", errors.Errorf("Timed out waiting for the LB %s to become ready", lbId)
+			return "", errors.Wrapf(err, "Error waiting for LB %s", lbId)
 		}
 
 		return listenerObj.ID, nil
@@ -896,23 +877,19 @@ func getConfigMap(kubeClient client.Client, namespace, name string) (*v1.ConfigM
 }
 
 func getUserCACert(kubeClient client.Client) (string, error) {
-	cm := &v1.ConfigMap{
-		TypeMeta:   metav1.TypeMeta{APIVersion: "v1", Kind: "ConfigMap"},
-		ObjectMeta: metav1.ObjectMeta{Name: UserCABundleConfigMap},
-	}
-	err := kubeClient.Get(context.TODO(), client.ObjectKey{Namespace: OpenShiftConfigNamespace, Name: UserCABundleConfigMap}, cm)
+	cm, err := getConfigMap(kubeClient, OpenShiftConfigNamespace, UserCABundleConfigMap)
 	if err != nil {
 		return "", err
 	}
-	return string(cm.Data["ca-bundle.pem"]), nil
+	return cm.Data["ca-bundle.pem"], nil
 }
 
-func getSavedAnnotation(kubeClient client.Client, annotation string) (string, error) {
+func getSavedOctaviaProvider(kubeClient client.Client) (string, error) {
 	cm, err := getConfigMap(kubeClient, KuryrNamespace, KuryrConfigMapName)
 	if err != nil {
 		return "", err
 	}
-	return cm.Annotations[annotation], nil
+	return cm.Annotations[names.KuryrOctaviaProviderAnnotation], nil
 }
 
 // Logs into OpenStack and creates all the resources that are required to run
@@ -1103,26 +1080,14 @@ func BootstrapKuryr(conf *operv1.NetworkSpec, kubeClient client.Client) (*bootst
 	}
 
 	var sgRules = []sgRule{
-		{podSgId, "0.0.0.0/0", 0, 0, rules.ProtocolTCP},
-		{masterSgId, openStackSvcCIDR, etcdClientPort, etcdClientPort, rules.ProtocolTCP},
+		{podSgId, "0.0.0.0/0", -1, -1, rules.ProtocolTCP},
+		{masterSgId, openStackSvcCIDR, etcdPort, etcdPort, rules.ProtocolTCP},
 		{masterSgId, openStackSvcCIDR, apiPort, apiPort, rules.ProtocolTCP},
-		// NOTE (maysamacedo): Splitting etcd sg port ranges in different
-		// rules to avoid the issue of constant leader election changes
-		{masterSgId, workerSubnet.CIDR, etcdClientPort, etcdClientPort, rules.ProtocolTCP},
-		{masterSgId, workerSubnet.CIDR, etcdServerToServerPort, etcdServerToServerPort, rules.ProtocolTCP},
-	}
-
-	var decommissionedRules = []sgRule{
-		{podSgId, workerSubnet.CIDR, 0, 0, ""},
-		{masterSgId, openStackSvcCIDR, etcdClientPort, etcdServerToServerPort, rules.ProtocolTCP},
-		// NOTE(maysamacedo): This sg rule is created by the installer. We need to remove it and
-		// create two more, each with a unique port from the range.
-		{masterSgId, workerSubnet.CIDR, etcdClientPort, etcdServerToServerPort, rules.ProtocolTCP},
 	}
 
 	for _, cidr := range podSubnetCidrs {
 		sgRules = append(sgRules,
-			sgRule{masterSgId, cidr, etcdClientPort, etcdClientPort, rules.ProtocolTCP},
+			sgRule{masterSgId, cidr, etcdPort, etcdPort, rules.ProtocolTCP},
 			sgRule{masterSgId, cidr, dnsPort, dnsPort, rules.ProtocolTCP},
 			sgRule{masterSgId, cidr, dnsPort, dnsPort, rules.ProtocolUDP},
 			sgRule{workerSgId, cidr, dnsPort, dnsPort, rules.ProtocolTCP},
@@ -1140,11 +1105,6 @@ func BootstrapKuryr(conf *operv1.NetworkSpec, kubeClient client.Client) (*bootst
 			sgRule{workerSgId, cidr, 9000, 9999, rules.ProtocolTCP},
 			sgRule{workerSgId, cidr, 9000, 9999, rules.ProtocolUDP},
 		)
-
-		decommissionedRules = append(decommissionedRules,
-			sgRule{masterSgId, cidr, 0, 0, ""},
-			sgRule{workerSgId, cidr, 0, 0, ""},
-		)
 	}
 
 	log.Print("Allowing required traffic")
@@ -1157,33 +1117,6 @@ func BootstrapKuryr(conf *operv1.NetworkSpec, kubeClient client.Client) (*bootst
 	}
 	log.Print("All requried traffic allowed")
 
-	// It may happen that we tightened some SG rules in an upgrade, we need to make sure to remove the ones that are
-	// not expected anymore.
-	log.Print("Removing old SG rules")
-	existingRules, err := listOpenStackSgRules(client, tag)
-	if err != nil {
-		return nil, errors.Wrap(err, "failed to list SG rules")
-	}
-
-	for _, existingRule := range existingRules {
-		// Need to convert to "our" format for easy comparisons.
-		rule := sgRule{existingRule.SecGroupID, existingRule.RemoteIPPrefix, existingRule.PortRangeMin,
-			existingRule.PortRangeMax, rules.RuleProtocol(existingRule.Protocol)}
-		for _, decommissionedRule := range decommissionedRules {
-			if decommissionedRule == rule {
-				log.Printf("Removing decommisioned rule %s (%s, %d, %d, %s) from SG %s", existingRule.ID,
-					existingRule.RemoteIPPrefix, existingRule.PortRangeMin, existingRule.PortRangeMax,
-					existingRule.Protocol, existingRule.SecGroupID)
-				err = rules.Delete(client, existingRule.ID).ExtractErr()
-				if err != nil {
-					return nil, errors.Wrapf(err, "Could not delete SG rule %s", existingRule.ID)
-				}
-				break
-			}
-		}
-	}
-	log.Print("All old SG rules removed")
-
 	lbClient, err := openstack.NewLoadBalancerV2(provider, gophercloud.EndpointOpts{})
 	if err != nil {
 		return nil, errors.Wrap(err, "failed to create Octavia client")
@@ -1228,11 +1161,9 @@ func BootstrapKuryr(conf *operv1.NetworkSpec, kubeClient client.Client) (*bootst
 	log.Print("Creating OpenShift API loadbalancer pool members")
 	r, _ := regexp.Compile(fmt.Sprintf("^%s-(master-port-[0-9]+|bootstrap-port)$", clusterID))
 	portList, err := listOpenStackPortsMatchingPattern(client, tag, r)
-	addresses := make([]string, 0)
 	for _, port := range portList {
 		if len(port.FixedIPs) > 0 {
 			portIp := port.FixedIPs[0].IPAddress
-			addresses = append(addresses, portIp)
 			log.Printf("Found port %s with IP %s", port.ID, portIp)
 
 			// We want bootstrap to stop being used as soon as possible, as it will serve
@@ -1254,11 +1185,6 @@ func BootstrapKuryr(conf *operv1.NetworkSpec, kubeClient client.Client) (*bootst
 		}
 	}
 
-	err = purgeOpenStackLbPoolMember(lbClient, poolId, addresses)
-	if err != nil {
-		return nil, errors.Wrap(err, "Failed on purging invalid LB members from LB pool")
-	}
-
 	log.Print("Ensuring certificates")
 	ca, key, err := ensureCA(kubeClient)
 	if err != nil {
@@ -1269,44 +1195,57 @@ func BootstrapKuryr(conf *operv1.NetworkSpec, kubeClient client.Client) (*bootst
 		return nil, errors.Wrap(err, "failed to ensure Certificate")
 	}
 
+	log.Print("Checking OVN Octavia driver support")
+	octaviaProviderSupport, err := IsOctaviaVersionSupported(lbClient, MinOctaviaVersionWithProviders)
+	if err != nil {
+		return nil, errors.Wrap(err, "failed to determine if Octavia supports providers")
+	}
+
 	log.Print("Checking Double Listeners Octavia support")
 	octaviaMultipleListenersSupport, err := IsOctaviaVersionSupported(lbClient, MinOctaviaVersionWithMultipleListeners)
 	if err != nil {
 		return nil, errors.Wrap(err, "failed to determine if Octavia supports double listeners")
 	}
 
-	octaviaVersion, err := getSavedAnnotation(kubeClient, names.KuryrOctaviaVersionAnnotation)
-	if err != nil && !apierrors.IsNotFound(err) {
-		return nil, errors.Wrap(err, "failed to get kuryr-config ConfigMap")
-	}
-
-	maxOctaviaVersion, err := getMaxOctaviaAPIVersion(lbClient)
-	if err != nil {
-		return nil, errors.Wrap(err, "failed to get max octavia api version")
-	}
-
-	// In case the Kuryr config-map is annotated with an Octavia version different
-	// than the current Octavia version, and older than the version that multiple
-	// listeners becomes available and the Octavia provider is amphora, an Octavia
-	// upgrade happened and UDP listeners are now allowed to be created.
-	// By recreating the OpenShift DNS service a new load balancer amphora is in
-	// place with all required listeners.
-	log.Print("Checking Octavia upgrade happened")
-	if octaviaVersion != "" {
-		savedOctaviaVersion := semver.MustParse(octaviaVersion)
-		multipleListenersVersion := semver.MustParse(MinOctaviaVersionWithMultipleListeners)
-		if !savedOctaviaVersion.Equal(maxOctaviaVersion) && savedOctaviaVersion.LessThan(multipleListenersVersion) && octaviaMultipleListenersSupport {
-			dnsService := &v1.Service{
-				TypeMeta:   metav1.TypeMeta{APIVersion: "v1", Kind: "Service"},
-				ObjectMeta: metav1.ObjectMeta{Name: DNSServiceName, Namespace: DNSNamespace},
-			}
-			err := kubeClient.Delete(context.TODO(), dnsService)
+	// Logic here is as follows:
+	// 1. We don't want to suddenly reconfigure Kuryr to use different provider, so we always try to fetch the currently
+	//    used one from annotation added to kuryr-config ConfigMap first.
+	// 2. If fetching annotation fails, then either it's the first run and we're yet to create the ConfigMap or user
+	//    deleted the annotation in order for us to trigger the reconfiguration. In both cases proceed with detection:
+	//    a. Check if Octavia version supports provider discovery.
+	//    b. List providers and look for OVN one.
+	//    c. If it's present configure Kuryr to use it.
+	//    d. In case of any issues just use whatever the default is.
+	octaviaProvider, err := getSavedOctaviaProvider(kubeClient) // Ignore error, just do the normal discovery then.
+	if octaviaProvider != "" {
+		log.Printf("Detected that Kuryr was already configured to use %s LB provider. Making sure to keep it that way.",
+			octaviaProvider)
+		if octaviaProvider == OVNProvider {
+			// OVN provider does not support multiple listeners, so in that case we always set false to that.
+			octaviaMultipleListenersSupport = false
+		}
+	} else {
+		octaviaProvider = "default"
+		if octaviaProviderSupport {
+			page, err := providers.List(lbClient, providers.ListOpts{}).AllPages()
 			if err != nil {
-				return nil, errors.Wrapf(err, "Failed to delete %s Service", DNSServiceName)
+				log.Print("failed to get lbs provider list, using default octavia provider")
+			} else {
+				providers, err := providers.ExtractProviders(page)
+				if err != nil {
+					return nil, errors.Wrap(err, "failed to extract providers list")
+				}
+				for _, provider := range providers {
+					if provider.Name == OVNProvider {
+						log.Print("OVN Provider is enabled and Kuryr will use it")
+						octaviaProvider = OVNProvider
+						// OVN provider does not support multiple listeners, so in that case we always set false to that.
+						octaviaMultipleListenersSupport = false
+					}
+				}
 			}
 		}
 	}
-	octaviaVersion = maxOctaviaVersion.Original()
 
 	log.Print("Kuryr bootstrap finished")
 
@@ -1319,9 +1258,9 @@ func BootstrapKuryr(conf *operv1.NetworkSpec, kubeClient client.Client) (*bootst
 			PodSecurityGroups:        []string{podSgId},
 			ExternalNetwork:          externalNetwork,
 			ClusterID:                clusterID,
+			OctaviaProvider:          octaviaProvider,
 			OctaviaMultipleListeners: octaviaMultipleListenersSupport,
 			OpenStackCloud:           cloud,
-			OctaviaVersion:           octaviaVersion,
 			WebhookCA:                b64.StdEncoding.EncodeToString(ca),
 			WebhookCAKey:             b64.StdEncoding.EncodeToString(key),
 			WebhookKey:               b64.StdEncoding.EncodeToString(webhookKey),
